# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from contextlib import closing
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import duckdb
from vdk.api.plugin.plugin_input import PEP249Connection
from vdk.internal.builtin_plugins.ingestion.ingester_base import IIngesterPlugin
from vdk.internal.core import errors
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.duckdb.duckdb_configuration import DuckDBConfiguration

log = logging.getLogger(__name__)


class IngestToDuckDB(IIngesterPlugin):
    """
    Create a new ingestion mechanism for ingesting to a DuckDB database
    """

    def __init__(
        self,
        conf: DuckDBConfiguration,
        new_connection_func: Callable[[], PEP249Connection],
    ):
        self._new_connection_func = new_connection_func
        self._conf = conf

    def ingest_payload(
        self,
        payload: List[Dict[str, Any]],
        destination_table: Optional[str] = None,
        target: str = None,
        collection_id: Optional[str] = None,
        metadata: Optional[IIngesterPlugin.IngestionMetadata] = None,
    ) -> None:
        """
        Performs the ingestion
        """
        if not payload:
            log.debug(
                f"Payload is empty. "
                f"Nothing to ingest into {target}, table {destination_table} and collection_id: {collection_id}"
            )
            return

        log.info(
            f"Ingesting payloads for target: {target}; "
            f"collection_id: {collection_id}"
        )

        with closing(self._new_connection_func().cursor()) as cur:
            if self._conf.get_auto_create_table_enabled():
                self.__create_table_if_not_exists(cur, destination_table, payload)
            else:
                self.__check_destination_table_exists(destination_table, cur)
            self.__ingest_payload(destination_table, payload, cur)

    def __ingest_payload(
        self, destination_table: str, payload: List[dict], cur: duckdb.cursor
    ) -> None:
        # Start a new transaction
        cur.execute("BEGIN TRANSACTION")

        try:
            keys = payload[0].keys()
            values = [[dic[k] for k in keys] for dic in payload]

            placeholders = ", ".join(["?" for _ in keys])
            sql = f"INSERT INTO {destination_table} ({', '.join(keys)}) VALUES ({placeholders})"

            cur.executemany(sql, values)

            cur.execute("COMMIT")
        except Exception:
            cur.execute("ROLLBACK")
            raise

    def __check_destination_table_exists(
        self, destination_table: str, cur: duckdb.cursor
    ) -> None:
        if not self._check_if_table_exists(destination_table, cur):
            errors.report_and_throw(
                UserCodeError(
                    "Cannot send payload for ingestion to DuckDB database.",
                    "destination_table does not exist in the target database.",
                    "Will not be able to send the payloads and will throw exception."
                    "Likely the job would fail",
                    "Make sure the destination_table exists in the target DuckDB database.",
                )
            )

    @staticmethod
    def _check_if_table_exists(table_name: str, cur: duckdb.cursor) -> bool:
        cur.execute("show tables")
        tables = cur.fetchall()
        return (table_name,) in tables

    def __table_columns(
        self, cur: duckdb.cursor, destination_table: str
    ) -> List[Tuple[str, str]]:
        columns = []
        if self._check_if_table_exists(destination_table, cur):
            cur.execute(
                f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{destination_table}'"
            )
            for row in cur.fetchall():
                columns.append((row[0], row[1]))
        return columns

    def __create_table_if_not_exists(
        self, cur: duckdb.cursor, destination_table: str, payload: List[dict]
    ):
        if not self._check_if_table_exists(destination_table, cur):
            log.info(
                f"Table {destination_table} does not exists. "
                f"Will auto-create it now based on first batch of input data."
            )
            columns = self.__infer_columns_from_payload(payload)
            self.__create_table(cur, destination_table, columns)
            log.info(f"Table {destination_table} created.")

    @staticmethod
    def __create_table(
        cur: duckdb.cursor, destination_table: str, columns: Dict[str, str]
    ):
        names = [
            f"{col_name} {col_type}"
            if " " not in col_name
            else f'"{col_name}" {col_type}'
            for col_name, col_type in columns.items()
        ]
        columns_as_sql_expression = ",".join(names)
        sql = f"CREATE TABLE IF NOT EXISTS {destination_table} ({columns_as_sql_expression})"
        log.debug(f"Create table using {sql}")
        cur.execute(sql)

    def __infer_columns_from_payload(self, payload: List[Dict]):
        columns = dict()
        for row in payload:
            for col, val in row.items():
                if col not in columns:
                    columns[col] = self.__python_value_to_duckdb_type(val)
                elif columns[col] == "NULL":
                    columns[col] = self.__python_value_to_duckdb_type(val)
        columns = {
            col: typ if typ != "NULL" else "VARCHAR" for col, typ in columns.items()
        }
        return columns

    @staticmethod
    def __python_value_to_duckdb_type(value: Any) -> str:
        if value is None:
            return "NULL"
        if isinstance(value, int):
            return "INTEGER"
        if isinstance(value, float):
            return "FLOAT"
        if isinstance(value, bool):
            return "BOOLEAN"
        if isinstance(value, str):
            return "VARCHAR"
        return "VARCHAR"



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import click
import duckdb
from tabulate import tabulate
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.util.decorators import closing_noexcept_on_close
from vdk.plugin.duckdb import duckdb_configuration
from vdk.plugin.duckdb.duckdb_configuration import DUCKDB_CONFIGURATION_DICTIONARY
from vdk.plugin.duckdb.duckdb_configuration import DUCKDB_DATABASE
from vdk.plugin.duckdb.duckdb_configuration import DuckDBConfiguration
from vdk.plugin.duckdb.ingest_to_duckdb import IngestToDuckDB

log = logging.getLogger(__name__)
"""
Include the plugins implementation. For example:
"""


@hookimpl
def vdk_configure(config_builder: ConfigurationBuilder) -> None:
    """Define the configuration settings needed for duckdb"""
    duckdb_configuration.add_definitions(config_builder)


@hookimpl
def initialize_job(context: JobContext) -> None:
    conf = DuckDBConfiguration(context.core_context.configuration)

    context.connections.add_open_connection_factory_method(
        "DUCKDB", lambda: duckdb.connect(conf.get_duckdb_database())
    )

    context.ingester.add_ingester_factory_method(
        "duckdb",
        (
            lambda: IngestToDuckDB(
                conf, lambda: context.connections.open_connection("DUCKDB")
            )
        ),
    )


@click.command(
    name="duckdb-query", help="Execute a DuckDB query against a local DUCKDB database."
)
@click.option("-q", "--query", type=click.STRING, required=True)
@click.pass_context
def duckdb_query(ctx: click.Context, query):
    conf = ctx.obj.configuration
    duckdb_db = conf.get_value(DUCKDB_DATABASE)
    conn = duckdb.connect(database=duckdb_db)

    with closing_noexcept_on_close(conn.cursor()) as cursor:
        cursor.execute(query)
        column_names = (
            [column_info[0] for column_info in cursor.description]
            if cursor.description
            else ()  # same as the default value for the headers parameters of the tabulate function
        )
        res = cursor.fetchall()
        click.echo(tabulate(res, headers=column_names))


@hookimpl
def vdk_command_line(root_command: click.Group):
    """Here we extend the vdk with a new command called "duckdb-query"
    enabling users to execute"""
    root_command.add_command(duckdb_query)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import os
from unittest import mock

import duckdb
import pytest
from click.testing import CliRunner
from click.testing import Result
from vdk.plugin.duckdb import duckdb_plugin
from vdk.plugin.test_utils.util_funcs import cli_assert_equal
from vdk.plugin.test_utils.util_funcs import CliEntryBasedTestRunner
from vdk.plugin.test_utils.util_funcs import jobs_path_from_caller_directory


@pytest.fixture(scope="function")
def setup_duckdb(tmpdir):
    temp_db_file = os.path.join(str(tmpdir), "test_db.duckdb")
    connection = duckdb.connect(temp_db_file)

    with mock.patch.dict(
        os.environ,
        {
            "DB_DEFAULT_TYPE": "duckdb",
            "DUCKDB_DATABASE": temp_db_file,
            "DUCKDB_INGEST_AUTO_CREATE_TABLE_ENABLED": "true",
        },
    ):
        yield

        connection.close()
        os.remove(temp_db_file)


def test_duckbd_plugin(setup_duckdb):
    runner = CliEntryBasedTestRunner(duckdb_plugin)

    result: Result = runner.invoke(
        ["run", jobs_path_from_caller_directory("test-duckdb-job")]
    )

    cli_assert_equal(0, result)

    _verify_sql_steps(runner)
    _verify_ingest_step(runner)


def _verify_sql_steps(runner):
    actual_rs = _sql_query(runner, "SELECT * FROM stocks")
    cli_assert_equal(0, actual_rs)
    expected_data = [
        {"date": "2020-01-01", "symbol": "GOOG", "price": 123.0},
        {"date": "2020-01-01", "symbol": "GOOG", "price": 123.0},
    ]
    assert json.loads(actual_rs.output) == expected_data


def _verify_ingest_step(runner):
    actual_rs = _sql_query(runner, "SELECT * FROM test_duckdb_table")
    cli_assert_equal(0, actual_rs)
    expected_data = [{'bool_col': 0, 'dec_col': '1.234', 'int_col': 2, 'str_col': 'str'}]
    assert json.loads(actual_rs.output) == expected_data


def _sql_query(runner, query):
    actual_rs: Result = runner.invoke(
        ["sql-query", "-o", "json", "--query", query],
        runner=CliRunner(
            mix_stderr=False
        ),  # TODO: replace when CliEntryBasedTestRunner add support for it
    )
    return actual_rs



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import pathlib
import tempfile
from typing import Dict
from typing import Optional

from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

DUCKDB_DATABASE = "DUCKDB_DATABASE"
DUCKDB_INGEST_AUTO_CREATE_TABLE_ENABLED = "DUCKDB_INGEST_AUTO_CREATE_TABLE_ENABLED"
DUCKDB_CONFIGURATION_DICTIONARY = "DUCKDB_CONFIGURATION_DICTIONARY"


class DuckDBConfiguration:
    def __init__(self, configuration: Configuration):
        self.__config = configuration

    def get_auto_create_table_enabled(self) -> bool:
        return self.__config.get_value(DUCKDB_INGEST_AUTO_CREATE_TABLE_ENABLED)

    def get_duckdb_database(self):
        return self.__config.get_value(DUCKDB_DATABASE) or "default_path.duckdb"

    def get_duckdb_configuration_dictionary(self) -> Optional[Dict[str, str]]:
        config_dict_str = self.__config.get_value(DUCKDB_CONFIGURATION_DICTIONARY)
        if config_dict_str:
            return json.loads(config_dict_str)
        else:
            return None


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=DUCKDB_DATABASE,
        default_value=str(
            pathlib.Path(tempfile.gettempdir()).joinpath("vdk-duckdb.db")
        ),
        description="The file of the DuckDB database.",
    )
    config_builder.add(
        key=DUCKDB_INGEST_AUTO_CREATE_TABLE_ENABLED,
        default_value=True,
        description="If set to true, auto create table if it does not exist during ingestion."
        "This is only applicable when ingesting data into DuckDB (ingest method is DuckDB).",
    )
    config_builder.add(
        key=DUCKDB_CONFIGURATION_DICTIONARY,
        default_value=None,
        description="A valid json string with config dictionary of duckdb configuration."
        " Those are configuration options set by https://duckdb.org/docs/sql/configuration.html",
    )
